import pandas as pd
from loguru import logger
from midiutil import MIDIFile

from .data import data_standardize as _pitch_data_standardize
from .data import shape_validator as _shape_validator
from .mapper import LinearMinMaxMapper


def build_track(pitches, midi_object, track, **params):
    """
    build_track builds the specified track for the midi object

    :param pitches: specified series of pitch for the track
    :type pitches: Union[pd.Series, np.ndarray, list, tuple]
    :param midi_object: instantiated midi object
    :param track: specified track id
    :type track: int
    """
    channel = params.get("channel", 0)
    duration = params.get("duration", 2)
    begin_beat = params.get("begin_beat", 0)  # In beats
    volume = params.get("volume", 127)  # 0-127, as per the MIDI standard

    for t, pitch in enumerate(pitches):
        midi_object.addNote(
            track, channel, pitch, begin_beat + t * duration, duration, volume
        )


def audiolizer(data, target, **params):
    """
    audiolize audiolizes the input data

    :param data: input tabular data to be audiolized
    :type data: Union[pd.DataFrame, pd.Series, ndarray, list, tuple]
    """
    midi_data = data.copy()

    # get params
    midi_object = params.get("midi_object")
    channel = params.get("channel", 0)
    begin_beat = params.get("begin_beat", 0)  # In beats
    duration = params.get("duration", 2)  # In beats
    tempo = params.get("tempo", 200)  # In BPM
    volume = params.get("volume", 127)  # 0-127, as per the MIDI standard

    # set mapper params
    mapper = params.get("mapper")
    pitch_min = params.get("pitch_min")
    pitch_max = params.get("pitch_max")

    # determine track names if the input data is a dataframe
    pitch_columns = params.get("pitch_columns")
    track_names = params.get("track_names")
    if track_names is None:
        if pitch_columns is None and isinstance(midi_data, pd.DataFrame):
            track_names = midi_data.columns.tolist()
            logger.debug(f"Track names:\n{track_names}")
        elif pitch_columns is not None:
            track_names = pitch_columns

    # standardize data
    if pitch_columns:
        pitch_raw_data = midi_data[pitch_columns]
    else:
        pitch_raw_data = midi_data.copy()
    pitch_raw_data = _pitch_data_standardize(pitch_raw_data)
    all_track_pitches = pitch_raw_data.get("pitches")
    n_tracks = pitch_raw_data.get("n_tracks")
    track_max = pitch_raw_data.get("max")
    track_min = pitch_raw_data.get("min")

    # instantiate midifile object
    if midi_object is None:
        midi_object = MIDIFile(n_tracks)

    if mapper is None:
        mapper = LinearMinMaxMapper(
            pitch_min=pitch_min,
            pitch_max=pitch_max,
            data_min=min(track_min),
            data_max=max(track_max),
        )

    for track in range(n_tracks):
        # add tempo
        midi_object.addTempo(track, 0, tempo)

        # add track names if track names are given
        if track_names:
            midi_object.addTrackName(track, 0, track_names[track])

    for track, pitches in enumerate(all_track_pitches):
        pitches = mapper.map(pitches)

        build_track(
            pitches,
            midi_object,
            track,
            channel=channel,
            duration=duration,
            begin_beat=begin_beat,
            volume=volume,
        )

    # export midi file
    with open(target, "wb") as output_file:
        midi_object.writeFile(output_file)



import os
import sys
import pandas as pd

import click
from loguru import logger
from audiorepr import audiolize

# determine current working directory
__CWD__ = os.getcwd()


@click.group(invoke_without_command=True)
@click.pass_context
def audiorepr(ctx):
    """
    audiorepr is the root command line command
    """
    if ctx.invoked_subcommand is None:
        click.echo("Hello {}".format(os.environ.get("USER", "")))
        click.echo("Welcome to audiorepr.")
    else:
        click.echo("Loading Service: %s" % ctx.invoked_subcommand)

@audiorepr.command()
def help():
    """
    How to use command line tool of audiorepr
    """

    click.echo(
        "The audiorepr command creates audio files out of a tabular data file.\n"
        "`audiorepr help` shows this message.\n"
        "`audiorepr create` command is the core command to create audio."
    )

@audiorepr.command()
@click.option(
    '-d', '--data',
    type=str,
    prompt=f"Please specify your data file\n"
    f"Your current working directory is {__CWD__}\n"
    f"Use absolute path or path relative to {__CWD__}.\n"
)
@click.option(
    '-t', '--target',
    type=click.Path(),
    prompt=f"Please specify your output file\n"
    f"Your current working directory is {__CWD__}\n"
    f"Use absolute path or path relative to {__CWD__}.\n"
    f"URL is also possible."
)
@click.option(
    '-c', '--column',
    multiple=True, default=None
)
def create(data, target, column):
    """
    creates audio from data file
    """

    click.echo(
        f"Using {data} as data source."
    )
    if column is not None:
        column = list(column)
        click.echo(
            f"Using columns:\n{column}"
        )

    df = pd.read_csv(data)
    click.echo(
        f"Data loaded"
    )
    audiolize.audiolizer(df, target=target, pitch_columns=column)
    click.echo(
        f"Audio file has been saved as {target}"
    )



if __name__ == "__main__":

    pass



import numpy as np
import pandas as pd
from loguru import logger


def shape_validator(data):
    """
    shape_validator validates the shape of the input data

    :param data: numpy array to be validated
    :type data: np.ndarray
    """

    shape = data.shape

    if len(shape) > 2:
        raise ValueError(f"input data:\n{data}\nshould have rank 1 or 2.")
    elif len(shape) == 2 and shape[-1] > 12:
        raise ValueError(
            f"input data:\n{data}\nshould have less or equal to 12 columns."
        )
    elif len(shape) == 1:
        data = data.reshape(len(data), 1)

    return data.shape


def data_standardize(data):
    """
    _data_standardize standardize data input

    :param data: Input data
    :type data: Union[pd.DataFrame, pd.Series, np.ndarray, list, tuple]
    """

    if isinstance(data, (pd.DataFrame, pd.Series)):
        data = data.to_numpy()
    elif isinstance(data, np.ndarray):
        data = data
    elif isinstance(data, (list, tuple)):
        data = np.array(data)
    else:
        raise TypeError(
            f"Input data\n {data} \nshould be one of the following: DataFrame, Series, ndarray, list, tuple"
        )

    shape = shape_validator(data)
    n_tracks = shape[-1]

    res = {
        "n_tracks": n_tracks,
        "pitches": np.transpose(data),
        "max": np.amax(data, axis=1),
        "min": np.amin(data, axis=1),
    }
    logger.debug(f"number of tracks: {n_tracks}")

    return res



class BaseMapper:
    def __init__(
        self, pitch_min=None, pitch_max=None, data_min=None, data_max=None, **params
    ):
        self.name = "basemapper"
        if pitch_min is None:
            self.pitch_min = 16
        if pitch_max is None:
            self.pitch_max = 96

        self.data_min = data_min
        self.data_max = data_max

    def _mapper(self, data, **params):

        raise NotImplementedError("Please implement this method")

    def map(self, data, **params):

        return self._mapper(data, **params)


class LinearMinMaxMapper(BaseMapper):
    def __init__(
        self, pitch_min=None, pitch_max=None, data_min=None, data_max=None, **params
    ):
        super().__init__(
            pitch_min=pitch_min,
            pitch_max=pitch_max,
            data_min=data_min,
            data_max=data_max,
            **params
        )
        self.name = "LinearMinMaxMapper"

    def _transformed_pitch(self, x, data_max, data_min):

        return int(
            self.pitch_min
            + (x - data_min) * (self.pitch_max - self.pitch_min) / (data_max - data_min)
        )

    def _mapper(self, data, **params):

        data_min = params.get("data_min", self.data_min)
        data_max = params.get("data_max", self.data_max)

        if data_min is None:
            data_min = min(data)
        if data_max is None:
            data_max = max(data)

        res = [self._transformed_pitch(i, data_max, data_min) for i in data]

        return res
